/*
Algorithme de simulation du modèle d'Ising 2D
*/

package ising;

import maillage.*;
import java.util.Vector;
import java.util.Random;
import java.lang.Math;
import java.lang.Object;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.*;
import java.awt.Color;

import simu.*;



/**
 * Classe executant la simulation du mod&egravele d'Ising
 */
public class Ising extends SimuApplet 
{
    
    /*
    *****************************
    DECLARATION DES VARIABLES
    *****************************
    */

    //Paramètres du système
    int    nlignes;             //Nombre de lignes pour les grilles carrée et hexagonale
    int    ncolonnes;           //Nombre de colonnes pour les grilles carrée et hexagonale
  //int    npoints;             //Nombre de points utilisés pour le maillage de Delaunay
    double j_int =1;            //Constante d'interaction fixée à -1 (système antiferromagnétique)
    int    rayon =1;            //Rayon d'influence des voisins fixé à 1
    double temperature;         //Temperature
    double tc;                  //Température critique
    double e_tot = 0;           //Energie totale du système
    double e_moy = 0;           //Energie moyenne du système
    double a_tot = 0;           //Aimantation totale du système
    double a_moy = 0;           //Aimantation moyenne du système
    double dt;                  //DeltaT, pas pour la variation de température
    double h;                   //Valeur du champ magnétique extérieur

    //Paramètres de la grille
    Grille grid;                //Grille
    Random r = new Random();    //Nombre aléatoire pour choisir une cellule au hasard
    int    gamma;               //Grossiement


    //Variables de comptage
    int    nbmotif;             //Nombre de motifs pour la correlation
    int    nboucle;             //Nombre d'itérations a une T donnée, utilisé e_moy et a_moy
    int    nsousboucle;         //Nombre d'itérations de nboucle, permet de moyenner e_moy et a_moy
    int    ntest;               //Nombre d'itérations
    double nboucle_inter;       //Nombre d'itérations modulé par une expo pour simulation en mode calcul


    //Paramètres de la corrélation
    double[]   moyennecorrel;   //Valeur moyennes des coefficients de correlation par motif
    double[]   erreurcorrel;    //Erreur sur les valeurs moyennes des coefficients de corrélation
    double[][] coefcorrel;      //Coefficient de corrélation par motif non moyenné


    //Variables de l'affichage du réseau pour simulab
    Image    img;
    Graphics graph;
    String   gType; 
    int      typesimu;
        
    
    public Ising()
    {
        appletDoc("Ising_doc.xml");
    }


    /**
     * Initialisation des variables de l'Applet
     * @param in Liste des Valeurs donn&eacute;es par l'Applet
     **/
    public void initCalculLive(ListeValeurs in) throws SimuException
    {
        
        //Lecture des paramètres données par l'utilisateur dans l'applet Java
        nlignes       = in.lireEntier("nlignes");
        ncolonnes     = in.lireEntier("ncolonnes");
        gamma         = in.lireEntier("gamma");
        nboucle       = in.lireEntier("nboucle");
        nsousboucle   = in.lireEntier("nsousboucle");
        nboucle_inter = nboucle;
        dt            = in.lireDouble("dt");
        h             = in.lireDouble("hext");
        temperature   = in.lireDouble("temperature");

        ntest = 0; //Réinitialisation du nombre d'itérations        
        
        //Choix du type de grille
        gType = in.lireString("gtype");

        //Choix du type d'affichage
        Affichage strtypesimu = getAffichageChoisi();

        //Type d'affichage : calcul
        if(strtypesimu.label.equals("calcul"))
        {
            Affichage donneeaffich;
            donneeaffich         = getAffichage("imagecalcul");
            donneeaffich.imgdimx = nlignes*gamma;
            donneeaffich.imgdimy = ncolonnes*gamma;
            img                  = createImage(nlignes*gamma, ncolonnes*gamma);
            graph                = img.getGraphics();
            typesimu             = 1;   
        

            //Choix par l'utilisateur du type de grille
            gType = in.lireString("gtype");


            //Type de grille: carrée
            if(gType.equals("Grille carrée")){
                grid = new GrilleCarree(nlignes,ncolonnes,true);
                tc=2.3;    
            }

            //Type de grille: hexagonale
            if(gType.equals("Grille Hexagonale")){
                grid = new GrilleHexa(nlignes,ncolonnes,true);
                tc=3.6;
            }

            /*
			//Type de grille: Voronoi
            if(gType.equals("Grille de Voronoi")){
                grid = new GrilleVoronoi(nlignes,npoints);
            }
			*/
        }


        //Type d'affichage : Visuel
        else
        {
            Affichage donneeaffich;
            donneeaffich         = getAffichage("imagevisuelle");
            donneeaffich.imgdimx = nlignes*gamma;
            donneeaffich.imgdimy = ncolonnes*gamma;
            img                  = createImage(nlignes*gamma, ncolonnes*gamma);
            graph                = img.getGraphics();
            typesimu             = 0;

            //Choix par l'utilisateur du type de grille
            gType = in.lireString("gtype");


            //Type de grille: carrée
            if(gType.equals("Grille carrée")){
                grid = new GrilleCarree(nlignes,ncolonnes);
                tc=1.15;    
            }

            //Type de grille: hexagonale
            if(gType.equals("Grille Hexagonale")){
                grid = new GrilleHexa(nlignes,ncolonnes);
                tc=1.8;
            }

            /*
			//Type de grille: Voronoi
            if(gType.equals("Grille de Voronoi")){
                grid = new GrilleVoronoi(nlignes,npoints);
            }
			*/
        }



        //Recherche des voisins de tous les spins
        for(int i=0;i<nlignes*ncolonnes;i++)
        {
            grid.rechercheVoisins(i,rayon);
        }


        /*
        *********************************
        INITIALISATION DE LA CORRELATION
        *********************************
        */

        //Tableau de coefficients de correlation
        nbmotif=(ncolonnes/2 -1);
        coefcorrel  = new double[nbmotif][];
        int nbi     = 0;
        int nbj     = 0;

        for(int i=0;i<nbmotif;i++)
        {
            nbi           = ncolonnes/((i+1)*2+1);
            nbj           = nlignes/((i+1)*2+1);
            coefcorrel[i] = new double[nbi*nbj];
        }
        moyennecorrel = new double[nbmotif];
        erreurcorrel  = new double[nbmotif];

        //Definition de la limite du plot des coefficients de correlation
        Affichage donneeaffich2;
        donneeaffich2      = getAffichage("correlation");
        donneeaffich2.xfin = ""+(nbmotif*2+1);

    }



    /**
     * Effectue les calcul en live pour l'applet
     * @param out Liste des Valeurs &agrave; envoyer &agrave; l'Applet
     **/
    public ListeValeurs calculLive() throws SimuException
    {
        ListeValeurs out = new ListeValeurs();
        evolution_moyenne_grille();

        
        //Affichage des plots et de l'image du réseau de spin
        grid.conversionGraphics(graph,gamma,Color.blue,Color.green);    //Grossissement de l'image et couleur des spins        
        out.ajouter("grille",img);                                      //Affichage de la grille
        out.ajouterDouble("ns",ntest);                                  //Nombre d'itérations
        out.ajouterDouble("etot",e_moy);                                //Energie moyenne du système
        out.ajouterDouble("M",a_moy);                                   //aimantation moyenne du sysytème
        out.ajouterDouble("temperature",temperature);                   //Température du système


        //Affichage de type calcul : la température est incrémentée 
        if(typesimu == 1)
        {        
            temperature = temperature + dt;
        }


        //Affichage du plot des coefficients de correlation
        out.ajouterDouble("moyenne",1);
        out.ajouterDouble("nbmotif",1);
        for(int i=1;i<=nbmotif;i++)
        {
            out.ajouterDouble("moyenne",moyennecorrel[i-1]);
            out.ajouterDouble("nbmotif",i*2+1);
        }

        return out;
    }




    /**
     * Choix du type d'&eacute;volution de la grille &agrave; temp&eacute;rature fixe ou &agrave; temp&eacute;rature variable.
     * D&eacute;termine l'aimantation moyenne et l'&eacute;nergie moyenne dans le cas d'une temp&eacute;rature variable, ainsi que les coefficients de corr&eacute;lation.
     */
    public void evolution_moyenne_grille()
    {

        //Initialisation des valeurs moyennes de l'énergie et de l'aimantation
        e_moy=0;
        a_moy=0;
        double temp_aimantation;    //Valeur de l'aimantation recupérée via la fonction aimantation()
        double temp_energie;        //Valeur de l'énergie recupérée via la fonction energie_tot()

        //Si le mode choisi est le mode calcul
        if(typesimu == 1)
        {
            //Initialisation des valeurs totales de l'énergie et de l'aimantation
            e_tot=0;
            a_tot=0;

            //Initialisation des valeurs des tableaux utilisés pour le calcul de la corrélation
            for(int cpt=0;cpt<nbmotif ;cpt++)
            {
                moyennecorrel[cpt] = 0;
                erreurcorrel[cpt]  = 0;
            }

            //Sous boucle d'itération
            //On effectue nsousboucle fois la méthode "evolution_temportelle_grille" à une température fixée pour moyenner l'énergie totale et l'aimantation totale 
            for (int i=0; i<nsousboucle; i++)
            {
                //Evolution du système
                evolution_temporelle_grille(temperature);

                //Calcul les coefficents de correlation
                correlation();
                for(int cpt=0;cpt<nbmotif ;cpt++)
                {
                    for(int j=0;j<coefcorrel[cpt].length; j++)
                    {
                        moyennecorrel[cpt] = moyennecorrel[cpt]+coefcorrel[cpt][j];
                        erreurcorrel[cpt]  = erreurcorrel[cpt]+Math.pow(coefcorrel[cpt][j],2);
                    }
                }

                //Calcul de l'energie et l'aimantation totales du système à l'équilibre après nboucles itérations
                temp_energie     = energie_tot();
                temp_aimantation = aimantation();
                e_tot            = e_tot + temp_energie;
                a_tot            = a_tot + temp_aimantation;

            }

            //On fait la moyenne des coef de correlation obtenus avec les nsousboucles
            for(int cpt=0;cpt<nbmotif ;cpt++)
            {
                moyennecorrel[cpt] = moyennecorrel[cpt]/(coefcorrel[cpt].length*nsousboucle*1.0);
                erreurcorrel[cpt]  = erreurcorrel[cpt]-Math.pow(moyennecorrel[cpt],2);
            }

            //On fait la moyenne de l'énergie totale et de l'aimantation obtenues après les nsousboucles
            e_moy = e_tot/(nsousboucle*1.0); 
            a_moy = a_tot/(nsousboucle*1.0);

        }
        

        //Si le mode choisi est le mode visuel
        else
        {
                //Simple evolution temporelle de la grille et calcul des valeurs moyennes de l'énergie et de l'aimantation
                evolution_temporelle_grille(temperature);
                temp_energie     = energie_tot();
                temp_aimantation = aimantation();
                e_moy            = temp_energie;
                a_moy            = temp_aimantation;
        }

    }
    


    /**
     * Evolution temporelle d'une cellule pour une temp&eacute;rature donn&eacute;e
     * Flip du spin selon la m&eacute;thode de Monte-Carlo
     * @param temperature Temp&eacute;rature fix&eacute;e du syst&egrave;me
     */
    public void evolution_temporelle_grille(double temperature)
    {

        double  proba;      //Probabilité qu'un spin change d'état
        double  tirage;     //Résultat du tirage aléatoire pour savoir si le spin change d'état
        int     somme;      //Somme de la valeur des spins des voisins du spin testé
        double  ediff;      //Delta d'énergie si le spin testé est renversé
        int     nbvoisin;   //Nombre de voisins du spin testé
        Cellule cel;        //Spin testé
        int     etat_cel;   //Etat du spin testé
        int     idcel;      //Indice du spin testé


        //Si le mode choisi est le mode calcul
        if(typesimu == 1)
        {
            //Pour un champ exterieur nul (h=0)...
            if(h == 0.)
            {
                //...la courbe est très bruitée autour de la température critique : on multiplie le nombre d'itérations par un gaussienne
                nboucle = (int)nboucle_inter + (int)(nboucle_inter*Math.exp(-Math.pow((temperature - tc),2)/0.05));
            }
        }

        //En mode visuel, on réalise autant d'itérations qu'il y a de cellules
        else
        {
            nboucle = nlignes*ncolonnes;
        }


        //Boucle d'evolution du système (nombre de fois où va tester un spin avant de calculer l'énergie et l'aimantation moyenne du système)
        for (int boucle = 0; boucle < nboucle; boucle++) 
        {
            //Tirage aléatoire d'une cellule
            idcel = r.nextInt(nlignes*ncolonnes);

            somme=0;
            ediff=0;
            ntest++;

            //Recupération des paramètres des voisins et du nombre de voisins
            cel                     = grid.getCellule(idcel);   
            nbvoisin                = cel.nombreVoisins();                
            Vector<Cellule> voisins = cel.getVoisins();

            //Somme des spins des voisins
            for(int i=0;i<nbvoisin;i++)
            {
                //Etat du voisin 
                //voisinsget().statut() renvoie TRUE (+1) ou FALSE (-1)
                if(voisins.get(i).statut())
                {
                    somme=somme+1;
                }
                else
                {
                    somme=somme-1;
                }
            }

            //Recupération de la valeur du spin testé
            //cel.statut() renvoie TRUE (+1) ou FALSE (-1)
            if(cel.statut())
            {
                etat_cel=1; //
            }
            else
            {
                etat_cel=-1;
            }

            //calcul de la perte/gain d'énergie si le spin testé est inversé
            ediff=2.0*etat_cel*somme*j_int - etat_cel*h;

            //si perte d'énergie (ediff < 0) -> changement de l'état du spin
            if(ediff < 0 )
            {
                cel.changementStatut();
            }
            else  //sinon tirage au sort
            {
                tirage = r.nextDouble();                //nombre aléatoire compris entre 0 et 1            
                proba  = Math.exp(-ediff/temperature);   //proba d'un changement d'état du spin
                
                //si le tirage aléatoire est inférieur a la proba de changement alors il y a changement d'état                
                if(tirage <= proba)
                {
                    cel.changementStatut();
                }
            }
        }
    }


    /**
    * Calcule l'&eacute;nergie totale du syst&egrave;me
    * @return energie &Eacute;nergie totale du syst&egrave;me
    */   
    public double energie_tot()
    {
        double energie     = 0.0;   //Energie du système
        double energie_max = 0.0;   //Energie max du système
        int etat_cel       = 0;     //Etat de la cellule testée
        int somme          = 0;     //Somme des spins des voisins de la cellule testée
        int nbvoisin       = 0;      //Nombre de voisins de la cellule testée
        double e_mag       = 0.0;   //Energie magnétique du réseau dans le cas d'un champ externe

        Cellule cel;                //Cellule testée


        //calcul de la contribution de chaque spin avec ses voisins sur l'énergie totale du système
        for(int i=0; i<nlignes*ncolonnes ; i++)
        {
            cel=grid.getCellule(i);   
            nbvoisin=cel.nombreVoisins();
                
            Vector<Cellule> voisins=cel.getVoisins();
            
            //somme des spins des voisins        
            somme=0;

            for(int j=0;j<nbvoisin;j++)
            {
                //Etat du voisin (True = +1, False =-1)
                if(voisins.get(j).statut())
                {
                    somme=somme+1;
                }
                else
                {
                    somme=somme-1;
                }
            }
    
            //Récupération de la valeur du spin choisi
            //cel.statut() renvoie TRUE (+1) ou FALSE (-1)
            if(cel.statut())
            {
                etat_cel=1;
            }
            else
            {
                etat_cel=-1;
            }
            
            //Incrémentation de l'énergie totale par la contribution du spin testé
            energie = energie  - somme*etat_cel;
            //Incrémentation de l'énergie magnétique par la contribution du spin testé
            e_mag = e_mag + h*etat_cel;
        }
        
        //Récupération du nombre de voisin
        cel=grid.getCellule(1);   
        nbvoisin=cel.nombreVoisins();

        //Retrait des "double contribution" (A sur B et B sur A)
        energie = energie /2. + e_mag;
        //Calcul de l'énergie totale max
        energie_max = ( nbvoisin /2.)*nlignes * ncolonnes * (Math.abs(j_int)) + ( nlignes * ncolonnes * Math.abs(h) );
        //Normalisation de l'énergie totale du système par l'énergie totale max
        energie = energie / energie_max;

        return energie;  
    }
    
      
    /**
    * Calcule l'aimantation moyenne du syst&egrave;me
    * L'aimantation moyenne est la somme des valeurs de spin du syst&egrave;me divis&eacute;e par le nombre de spins
    * @return aimantation Valeur de l'aimantation du syst&egrave;me
    */
    public double aimantation()
    {
        double aimantation_tot = 0;              //aimantation totale du système
        double aimantation     = 0;              //aimantation

        //Somme de toutes les valeurs de spins
        for(int j=0; j<ncolonnes*nlignes; j++) 
        {
            if(grid.getCellule(j).statut())
            {
                aimantation_tot++;
            }
            else
            {
                aimantation_tot--;
            }
        }
        //Normalisation de l'aimantation totale par le nombre de spins pour obtenir l'aimantation
        aimantation=aimantation_tot/(ncolonnes*nlignes);
        return aimantation;
    }


    /**
    * Calcule le coefficient de corr&eacute;lation
    * la corr&eacute;lation donne une indication sur la valeur de spin des voisins proches. Si la corr&eacute;lation est de 1, tous les voisins sont de m&ecirc;me spin que la cellule consid&eacute;r&eacute;e, inversement si tout les voisins ont des spins diff&eacute;rents le coefficient de corr&eacute;lation tends vers 0.
    * La corr&eacute;lation est calcul&eacute;e dans une boite (sans recouvement) de taille croissante jusqu'&agrave; la taille maximale autoris&eacute;e (taille du syst&egrave;me pour la cellule centr&eacute;e)
    * @return coefcorrel Valeur finale des coefficients de corr&eacute;lation pour chaque taille de boite 
    */
    public double[][] correlation() 
    {
        //calcul des coef de correlation du système

        int i2             = 0;                                 //Indice de la deuxième dimention du tableau de coefcorrel (taille variable)
        double[][] cellule = new double[ncolonnes][nlignes];    //Tableau de la valeur des états de tout les spins
        double somme       = 0;                                 //Somme des états de tout les spins compris dans un motif
        double tmp         = 0;                                 //Variable de stockage
    
        //Recupération de toutes les valeurs de spins dans un tableau 2D 
        for (int i=0; i<ncolonnes; i++)
        {
            for(int j=0; j<nlignes; j++)
            {
                if(grid.getCellule(j).statut())
                    cellule[i][j]=1.0;
                else
                    cellule[i][j]=-1.0;
            }
        }

        //Boucle sur les tailles de motif, la taille augmente a chaque itération (taille d'un motif : i_motif*2+1 )
        for(int i_motif=1;i_motif<=nbmotif;i_motif++)
        {
            i2=0; // indice du 2eme tableau des coef de correl, correspond a la boite numero "i2+1" où est calculé le coef de correlation
            
            //boucle sur les colonnes, les boites se déplacent suivant cette colonne
            for(int j=i_motif;j<=ncolonnes-i_motif-1;j=j+i_motif*2+1)
            {
                //boucle sur les lignes, les boites se déplacent suivant cette ligne
                for(int k=i_motif;k<=nlignes-i_motif-1;k=k+i_motif*2+1)
                {
                    //double boucle (j2 et k2) pour sommer tous les pixels de la boite de taille 2*i_motif+1
                    for(int j2=j-i_motif;j2<=j+i_motif;j2++)
                    {
                        for(int k2=k-i_motif;k2<=k+i_motif;k2++)
                        {
                            //somme de toutes les valeurs des états des voisins du spin testé, plus la valeur du spin testé (qui sera retirée après)
                            somme=somme+cellule[j2][k2];
                        }
                }

                //calcul du coef de correlation en retirant la valeur du spin du centre (celui qu'on teste)
                tmp=(somme*cellule[j][k]-1.0)/(Math.pow((i_motif*2+1),2)-1);

                coefcorrel[i_motif-1][i2]=tmp;
                i2++;
                somme=0.0;
                }
            }
        }
        return coefcorrel;  
    }

}
